@@ -1,4 +1,4 @@
-// Copyright 2012, Google, Inc. All rights reserved.
+// Copyright 2012, 2018 GoPacket Authors. All rights reserved.
 //
 // Use of this source code is governed by a BSD-style license
 // that can be found in the LICENSE file in the root of the source
@@ -78,6 +78,48 @@ func TestParseDNSTypeTXT(t *testing.T) {
 	}
 }
 
+var testParseDNSTypeOPT = []byte{
+	0x00, 0x90, 0x0b, 0x12, 0x91, 0xc1, 0x00, 0x1c, 0xc0, 0x93, 0x33, 0xfb, 0x08, 0x00, 0x45, 0x00,
+	0x00, 0x5A, 0xce, 0x58, 0x00, 0x00, 0x40, 0x11, 0x67, 0xe2, 0xac, 0x10, 0x01, 0xc7, 0x4b, 0x4b,
+	0x4b, 0x4b, 0xd6, 0x00, 0x00, 0x35, 0x00, 0x46, 0x44, 0xb0, 0x50, 0x12, 0x01, 0x00, 0x00, 0x01,
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x03, 0x77, 0x77, 0x77, 0x04, 0x69, 0x65, 0x74, 0x66, 0x03,
+	0x6f, 0x72, 0x67, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x29, 0x10, 0x00, 0x00, 0x00, 0x80,
+	0x00, 0x00, 0x13, 0x69, 0x42, 0x00, 0x0F, 0x4F, 0x70, 0x65, 0x6E, 0x44, 0x4E, 0x53, 0x01, 0x23,
+	0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF,
+}
+
+func TestParseDNSTypeOPT(t *testing.T) {
+	p := gopacket.NewPacket(testParseDNSTypeOPT, LinkTypeEthernet, testDecodeOptions)
+
+	if p.ErrorLayer() != nil {
+		t.Error("Failed to decode packet:", p.ErrorLayer().Error())
+	}
+	checkLayers(p, []gopacket.LayerType{LayerTypeEthernet, LayerTypeIPv4, LayerTypeUDP, LayerTypeDNS}, t)
+	questions := p.Layer(LayerTypeDNS).(*DNS).Questions
+	if len(questions) != 1 {
+		t.Error("Failed to parse 1 DNS question")
+	}
+	additionals := p.Layer(LayerTypeDNS).(*DNS).Additionals
+	if len(additionals) != 1 {
+		t.Error("Failed to parse 1 DNS additional")
+	}
+
+	optAll := additionals[0].OPT
+	if len(optAll) != 1 {
+		t.Errorf("Parsed %d OPTs, expected 1", len(optAll))
+	}
+
+	if additionals[0].OPT[0].Code != DNSOptionCodeDeviceID {
+		t.Error("Failed to parse the OPT Code")
+	}
+	if string(additionals[0].OPT[0].Data[:7]) != "OpenDNS" {
+		t.Error("Failed to parse the Data Part 1")
+	}
+	if !bytes.Equal(additionals[0].OPT[0].Data[7:], []byte{0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF}) {
+		t.Error("Failed to parse the Data Part 2")
+	}
+}
+
 func testQuestionEqual(t *testing.T, i int, exp, got DNSQuestion) {
 	if !bytes.Equal(exp.Name, got.Name) {
 		t.Errorf("expected Questions[%d].Name = %v, got %v", i, string(exp.Name), string(got.Name))
@@ -173,6 +215,20 @@ func testResourceEqual(t *testing.T, i int, name string, exp, got DNSResourceRec
 	if exp.MX.Preference != got.MX.Preference {
 		t.Errorf("expected %s[%d].MX.Preference = %v, got %v", name, i, exp.MX.Preference, got.MX.Preference)
 	}
+
+	// OPT
+	if len(exp.OPT) != len(got.OPT) {
+		t.Errorf("expected len(%s[%d].OPT) = %v, got %v", name, i, len(exp.OPT), len(got.OPT))
+	}
+
+	for j := range exp.OPT {
+		if exp.OPT[j].Code != got.OPT[j].Code {
+			t.Errorf("expected %s[%d].OPT[%d].Code = %v, got %v", name, i, j, exp.OPT[j].Code, got.OPT[j].Code)
+		}
+		if !bytes.Equal(exp.OPT[j].Data, got.OPT[j].Data) {
+			t.Errorf("expected %s[%d].OPT[%d].Data = %v, got %v", name, i, j, exp.OPT[j].Data, got.OPT[j].Data)
+		}
+	}
 }
 
 func testDNSEqual(t *testing.T, exp, got *DNS) {
@@ -279,6 +335,51 @@ func TestDNSEncodeQuery(t *testing.T) {
 	testDNSEqual(t, dns, dns2)
 }
 
+func TestDNSEncodeQueryWithOPT(t *testing.T) {
+	dns := &DNS{ID: 1234, OpCode: DNSOpCodeQuery, RD: true}
+	dns.Questions = append(dns.Questions,
+		DNSQuestion{
+			Name:  []byte("example1.com"),
+			Type:  DNSTypeA,
+			Class: DNSClassIN,
+		})
+
+	dns.Questions = append(dns.Questions,
+		DNSQuestion{
+			Name:  []byte("example2.com"),
+			Type:  DNSTypeA,
+			Class: DNSClassIN,
+		})
+	dns.Additionals = append(dns.Additionals,
+		DNSResourceRecord{
+			Type:  DNSTypeOPT,
+			Class: 4096,
+			OPT: []DNSOPT{
+				DNSOPT{
+					Code: DNSOptionCodeDeviceID,
+					Data: []byte("OpenDNS"),
+				},
+			},
+		})
+
+	buf := gopacket.NewSerializeBuffer()
+	opts := gopacket.SerializeOptions{FixLengths: true}
+	err := gopacket.SerializeLayers(buf, opts, dns)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if int(dns.QDCount) != len(dns.Questions) {
+		t.Errorf("fix lengths did not adjust QDCount, expected %d got %d", len(dns.Questions), dns.QDCount)
+	}
+	if int(dns.ARCount) != len(dns.Additionals) {
+		t.Errorf("fix lengths did not adjust ARCount, expected %d got %d", len(dns.Additionals), dns.ARCount)
+	}
+
+	p2 := gopacket.NewPacket(buf.Bytes(), LayerTypeDNS, testDecodeOptions)
+	dns2 := p2.Layer(LayerTypeDNS).(*DNS)
+	testDNSEqual(t, dns, dns2)
+}
+
 func TestDNSEncodeResponse(t *testing.T) {
 	dns := &DNS{ID: 1234, QR: true, OpCode: DNSOpCodeQuery,
 		AA: true, RD: true, RA: true}
@@ -398,7 +499,7 @@ func TestDNSMalformedPacket2(t *testing.T) {
 	}
 }
 
-// testMalformedRootQuery is the packet:
+// testBlankNameRootQuery is the packet:
 //   08:31:18.143065 IP 10.77.0.26.53 > 10.1.0.233.65071: 59508- 0/13/3 (508)
 //   	0x0000:  0055 22af c637 0022 55ac deac 0800 4500  .U"..7."U.....E.
 //   	0x0010:  0218 76b2 4000 7211 7ad2 0a4d 001a 0a01  ..v.@.r.z..M....
@@ -435,7 +536,7 @@ func TestDNSMalformedPacket2(t *testing.T) {
 //   	0x0200:  0004 ca0c 1b21 c058 0001 0001 0000 0e10  .....!.X........
 //   	0x0210:  0004 c629 0004 c078 0001 0001 0000 0e10  ...)...x........
 //   	0x0220:  0004 c024 9411                           ...$..
-var testMalformedRootQuery = []byte{
+var testBlankNameRootQuery = []byte{
 	0x00, 0x55, 0x22, 0xaf, 0xc6, 0x37, 0x00, 0x22, 0x55, 0xac, 0xde, 0xac, 0x08, 0x00, 0x45, 0x00,
 	0x02, 0x18, 0x76, 0xb2, 0x40, 0x00, 0x72, 0x11, 0x7a, 0xd2, 0x0a, 0x4d, 0x00, 0x1a, 0x0a, 0x01,
 	0x00, 0xe9, 0x00, 0x35, 0xfe, 0x2f, 0x02, 0x04, 0xb8, 0xf5, 0xe8, 0x74, 0x81, 0x00, 0x00, 0x01,
@@ -473,12 +574,10 @@ var testMalformedRootQuery = []byte{
 	0x00, 0x04, 0xc0, 0x24, 0x94, 0x11,
 }
 
-func TestMalformedRootQuery(t *testing.T) {
-	p := gopacket.NewPacket(testMalformedRootQuery, LinkTypeEthernet, testDecodeOptions)
-	if errLayer := p.ErrorLayer(); errLayer == nil {
-		t.Error("No error layer on invalid DNS name")
-	} else if err := errLayer.Error(); !strings.Contains(err.Error(), "no dns data found") {
-		t.Errorf("unexpected error message: %v", err)
+func TestBlankNameRootQuery(t *testing.T) {
+	p := gopacket.NewPacket(testBlankNameRootQuery, LinkTypeEthernet, testDecodeOptions)
+	if err := p.ErrorLayer(); err != nil {
+		t.Error("Error layer on blank DNS name field:", err)
 	}
 }
 
@@ -748,6 +847,25 @@ func TestMalformedDNSOhGodMakeItStop(t *testing.T) {
 	}
 }
 
+var testMalformedDNSOPT = []byte{
+	0x00, 0x90, 0x0b, 0x12, 0x91, 0xc1, 0x00, 0x1c, 0xc0, 0x93, 0x33, 0xfb, 0x08, 0x00, 0x45, 0x00,
+	0x00, 0x5A, 0xce, 0x58, 0x00, 0x00, 0x40, 0x11, 0x67, 0xe2, 0xac, 0x10, 0x01, 0xc7, 0x4b, 0x4b,
+	0x4b, 0x4b, 0xd6, 0x00, 0x00, 0x35, 0x00, 0x46, 0x44, 0xb0, 0x50, 0x12, 0x01, 0x00, 0x00, 0x01,
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x03, 0x77, 0x77, 0x77, 0x04, 0x69, 0x65, 0x74, 0x66, 0x03,
+	0x6f, 0x72, 0x67, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x29, 0x10, 0x00, 0x00, 0x00, 0x80,
+	0x00, 0x00, 0x13, 0x69, 0x42, 0x00, 0x10, 0x4F, 0x70, 0x65, 0x6E, 0x44, 0x4E, 0x53, 0x01, 0x23,
+	0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF,
+}
+
+func TestMalformedDNSOPT(t *testing.T) {
+	p := gopacket.NewPacket(testMalformedDNSOPT, LinkTypeEthernet, testDecodeOptions)
+	if errLayer := p.ErrorLayer(); errLayer == nil {
+		t.Error("No error layer on invalid DNS name")
+	} else if err := errLayer.Error(); !strings.Contains(err.Error(), "Malformed DNSOPT record") {
+		t.Errorf("unexpected error message: %v", err)
+	}
+}
+
 // testPacketDNSPanic7 is the packet:
 //   07:56:25.174747 IP 10.77.0.11.53 > 10.1.0.67.55777: 41808*-| 3/7/0 TXT "google-site-verification=DC2uC-T8kD33lINhNzfo0bNBrw-vrCXs5BPF5BXY56g", TXT "v=spf1 include:spf-a.outlook.com include:spf-b.outlook.com ip4:157.55.9.128/25 include:spf.protection.outlook.com include:spf-a.hotmail.com include:_spf-ssg-b.microsoft.com include:_spf-ssg-c.microsoft.com ~all", TXT "google-site-verification=0iLWhIMhXEkeWwWfFU4ursTn-_OvoOjaA0Lr7Pg1sEM" (512)
 //   	0x0000:  0055 22af c637 0022 55ac deac 0800 4500  .U"..7."U.....E.
@@ -831,3 +949,40 @@ func TestPacketDNSPanic7(t *testing.T) {
 		t.Errorf("unexpected error message: %v", err)
 	}
 }
+
+func TestDNSPacketWriteAnswer(t *testing.T) {
+	dns := &DNS{ID: 0x1234, QR: true, OpCode: DNSOpCodeQuery, ResponseCode: DNSResponseCodeNoErr, Answers: []DNSResourceRecord{
+		DNSResourceRecord{
+			Name:  []byte("www.example.com"),
+			Type:  DNSTypeA,
+			Class: DNSClassIN,
+			IP:    net.IPv4(127, 0, 0, 1),
+		},
+		DNSResourceRecord{
+			Name:  []byte("www.example.com"),
+			Type:  DNSTypeAAAA,
+			Class: DNSClassIN,
+			IP:    net.IP{15: 1},
+		},
+	}}
+	buf := gopacket.NewSerializeBuffer()
+	opts := gopacket.SerializeOptions{ComputeChecksums: true, FixLengths: true}
+	if err := gopacket.SerializeLayers(buf, opts, dns); err != nil {
+		t.Fatal(err)
+	}
+	dns2 := &DNS{}
+	if err := dns2.DecodeFromBytes(buf.Bytes(), gopacket.NilDecodeFeedback); err != nil {
+		t.Fatalf("could not decode: %v", err)
+	}
+	if want, got := 2, len(dns2.Answers); want != got {
+		t.Fatalf("num answers, want %d got %d", want, got)
+	} else if got, want := string(dns2.Answers[0].Name), "www.example.com"; got != want {
+		t.Fatalf("unexpected first answer name %q, want %q", got, want)
+	} else if got, want := string(dns2.Answers[1].Name), "www.example.com"; got != want {
+		t.Fatalf("unexpected second answer name %q, want %q", got, want)
+	}
+	t.Log(gopacket.LayerString(dns2))
+	if want, got := 86, len(buf.Bytes()); want != got {
+		t.Fatalf("Encoded size, want %d got %d", want, got)
+	}
+}
